import netgen.meshing as ngmsh
import ngsolve as ngs

def get_mesh(phi, N, scale, offset, interp_ord=1, cutoff=-0.9954):
    # Mapping should only use mesh points, not the phi values for the 
    # additional quadrature points. Only H1 elements are supported.
    factor = interp_ord
    phi_idx = (phi*2.0 - 1.0)
    phi_idx = phi_idx[::factor,::factor]
        
    mesh = ngmsh.Mesh()
        
    points = []
    coords = []
    pts = []
    for i in range(N[0]+1):
        for j in range(N[1]+1):
            for k in range(N[2]+1):
                x = -offset[0] + scale[0]*i/N[0]
                y = -offset[1] + scale[1]*j/N[1]
                z = -offset[2] + scale[2]*k/N[2]
                points.append(mesh.Add(ngmsh.MeshPoint(ngmsh.Pnt(x,y,z))))
                coords.append((i,j, k))
                pts.append((x,y,z))
                   
    mesh.Add(ngmsh.FaceDescriptor(surfnr=1, domin=1, bc=1))
    mesh.Add(ngmsh.FaceDescriptor(surfnr=2, domin=2, bc=2))

    for i in range(N[0]):
        for j in range(N[1]):
            for k in range(N[2]):
                base = i*(N[1]+1)*(N[2]+1) + j*(N[2]+1) + k
                baseup = base + (N[1]+1)*(N[2]+1)
                p1 = base
                p2 = base + 1
                p3 = base + (N[2]+1) + 1
                p4 = base + (N[2]+1)
                p5 = baseup
                p6 = baseup + 1
                p7 = baseup + (N[2]+1) + 1
                p8 = baseup + (N[2]+1)
                idx = ((phi_idx[coords[p1]] < cutoff) or (phi_idx[coords[p2]] < cutoff) or (phi_idx[coords[p3]] < cutoff) or (phi_idx[coords[p4]] < cutoff) or (phi_idx[coords[p5]] < cutoff) or (phi_idx[coords[p6]] < cutoff) or (phi_idx[coords[p7]] < cutoff) or (phi_idx[coords[p8]] < cutoff)) + 1
                mesh.Add(ngmsh.Element3D(idx, [points[p1], points[p2], points[p3], points[p4], points[p5], points[p6], points[p7], points[p8]]))

    def add_bc(p, d, N, deta, neta, facenr):
    
        def add_seg(i, j, os):
            base = p + i*d + j*deta
            p1 = base
            p2 = base + os
            mesh.Add(ngmsh.Element1D([points[p1], points[p2]], index=facenr))
                
            return
        
        for i in range(N):
            for j in [0, neta]:
                add_seg(i, j, d)
            
        for i in [0, N]:
            for j in range(neta):
                add_seg(i, j, deta)
            
        for i in range(N):
            for j in range(neta):
                base = p + i*d + j*deta
                p1 = base
                p2 = base + d
                p3 = base + d + deta
                p4 = base + deta
                mesh.Add(ngmsh.Element2D(facenr, [points[p1], points[p2], points[p3], points[p4]]))
            
        return
            
    add_bc(0, 1, N[2], N[2]+1, N[1], 2)
    add_bc(0, (N[1]+1)*(N[2]+1), N[0], 1, N[2], 2)
    add_bc((N[0]+1)*(N[1]+1)*(N[2]+1)-1, -(N[2]+1), N[1], -1, N[2], 2)
    add_bc((N[0]+1)*(N[1]+1)*(N[2]+1)-1, -1, N[2], -(N[1]+1)*(N[2]+1), N[0], 2)
    add_bc(0, N[2]+1, N[1], (N[1]+1)*(N[2]+1), N[0], 2)    
    add_bc((N[0]+1)*(N[1]+1)*(N[2]+1)-1, -(N[1]+1)*(N[2]+1), N[0], -(N[2]+1), N[1], 2)

    mesh.SetMaterial(1, 'interior')
    mesh.SetMaterial(2, 'exterior')
    mesh.SetBCName(2,'exterior_bc')

    return ngs.Mesh(mesh)

def solve():
    N = [19, 19, 19]
    scale = [12.0, 12.0, 12.0]
    offset = [6.0, 6.0, 6.0]
    interp_ord = 2
    cutoff = -0.9954

    f = -12
    BC = {'Neumann': {},
          'Dirichlet': {'1': ngs.x*ngs.x + 2*ngs.y*ngs.y + 3*ngs.z*ngs.z},
          'Robin': {}}
    
    # phi is a numpy array with elements on [-1,1].
    phi = get_phi()
    
    # Manually construct the mesh. phi dictates which elements belong to which
    # material on the mesh.
    mesh = get_mesh(phi, N, scale, offset, interp_ord, cutoff)
    
    # Solely Dirichlet BCs requires using the Nitsche method to apply
    # the boundary conditions.
    beta = 10.0*interp_ord**2*min(N)/min(scale)
            
    fes_large = ngs.H1(mesh, order=interp_ord, definedon=mesh.Materials('interior'))
    fes = ngs.Compress(fes_large)
    
    # Convert phi into an NGSolve GridFunction defined over the mesh.
    p = numpy_to_NGSolve(phi)
    
    # Solve the Poisson problem. The Nitsche method is being used to implicitly 
    # impose the Dirichlet boundary condition. 
    u = fes.TrialFunction()
    v = fes.TestFunction()
            
    L = ngs.LinearForm(fes)
    L += ngs.SymbolicLFI(f*v*p, definedon=mesh.Materials('interior'))
    g = BC['Dirichlet']['1']
    L += ngs.SymbolicLFI(ngs.grad(v)*g*ngs.grad(p) + beta*g*v*ngs.grad(p)*ngs.grad(p), definedon=mesh.Materials('interior'))
    L.Assemble()
            
    a = ngs.BilinearForm(fes, symmetric=True)
    a += ngs.SymbolicBFI(ngs.grad(u)*ngs.grad(v)*p)
    a += ngs.SymbolicBFI(ngs.grad(u)*ngs.grad(p)*v + ngs.grad(v)*ngs.grad(p)*u + beta*u*v*ngs.grad(p)*ngs.grad(p))
    a.Assemble()
            
    gfu = ngs.GridFunction(fes)
    gfu.vec.data = a.mat.Inverse(fes.FreeDofs())*L.vec
            
    return gfu